from __future__ import annotations

import contextlib
import os
import pickle
import sys
import textwrap

import pytest

import env
import pybind11_tests

if env.IOS:
    pytest.skip("Subinterpreters not supported on iOS", allow_module_level=True)

# 3.14.0b3+, though sys.implementation.supports_isolated_interpreters is being added in b4
# Can be simplified when we drop support for the first three betas
CONCURRENT_INTERPRETERS_SUPPORT = (
    sys.version_info >= (3, 14)
    and (
        sys.version_info != (3, 14, 0, "beta", 1)
        and sys.version_info != (3, 14, 0, "beta", 2)
    )
    and (
        sys.version_info == (3, 14, 0, "beta", 3)
        or sys.implementation.supports_isolated_interpreters
    )
)


def get_interpreters(*, modern: bool):
    if modern and CONCURRENT_INTERPRETERS_SUPPORT:
        from concurrent import interpreters

        def create():
            return contextlib.closing(interpreters.create())

        def run_string(
            interp: interpreters.Interpreter,
            code: str,
            *,
            shared: dict[str, object] | None = None,
        ) -> Exception | None:
            if shared:
                interp.prepare_main(**shared)
            try:
                interp.exec(code)
                return None
            except interpreters.ExecutionFailed as err:
                return err

        return run_string, create

    if sys.version_info >= (3, 12):
        interpreters = pytest.importorskip(
            "_interpreters" if sys.version_info >= (3, 13) else "_xxsubinterpreters"
        )

        @contextlib.contextmanager
        def create(config: str = ""):
            try:
                if config:
                    interp = interpreters.create(config)
                else:
                    interp = interpreters.create()
            except TypeError:
                pytest.skip(f"interpreters module needs to support {config} config")

            try:
                yield interp
            finally:
                interpreters.destroy(interp)

        def run_string(
            interp: int, code: str, shared: dict[str, object] | None = None
        ) -> Exception | None:
            kwargs = {"shared": shared} if shared else {}
            return interpreters.run_string(interp, code, **kwargs)

        return run_string, create

    pytest.skip("Test requires the interpreters stdlib module")


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
def test_independent_subinterpreters():
    """Makes sure the internals object differs across independent subinterpreters"""

    sys.path.insert(0, os.path.dirname(pybind11_tests.__file__))

    run_string, create = get_interpreters(modern=True)

    import mod_per_interpreter_gil as m

    if not m.defined_PYBIND11_HAS_SUBINTERPRETER_SUPPORT:
        pytest.skip("Does not have subinterpreter support compiled in")

    code = textwrap.dedent(
        """
        import mod_per_interpreter_gil as m
        import pickle
        with open(pipeo, 'wb') as f:
            pickle.dump(m.internals_at(), f)
        """
    ).strip()

    with create() as interp1, create() as interp2:
        try:
            res0 = run_string(interp1, "import mod_shared_interpreter_gil")
            if res0 is not None:
                res0 = str(res0)
        except Exception as e:
            res0 = str(e)

        pipei, pipeo = os.pipe()
        run_string(interp1, code, shared={"pipeo": pipeo})
        with open(pipei, "rb") as f:
            res1 = pickle.load(f)

        pipei, pipeo = os.pipe()
        run_string(interp2, code, shared={"pipeo": pipeo})
        with open(pipei, "rb") as f:
            res2 = pickle.load(f)

    assert "does not support loading in subinterpreters" in res0, (
        "cannot use shared_gil in a default subinterpreter"
    )
    assert res1 != m.internals_at(), "internals should differ from main interpreter"
    assert res2 != m.internals_at(), "internals should differ from main interpreter"
    assert res1 != res2, "internals should differ between interpreters"


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
@pytest.mark.skipif(not CONCURRENT_INTERPRETERS_SUPPORT, reason="Requires 3.14.0b3+")
def test_independent_subinterpreters_modern():
    """Makes sure the internals object differs across independent subinterpreters. Modern (3.14+) syntax."""

    sys.path.insert(0, os.path.dirname(pybind11_tests.__file__))

    import mod_per_interpreter_gil as m

    if not m.defined_PYBIND11_HAS_SUBINTERPRETER_SUPPORT:
        pytest.skip("Does not have subinterpreter support compiled in")

    from concurrent import interpreters

    code = textwrap.dedent(
        """
        import mod_per_interpreter_gil as m

        values.put_nowait(m.internals_at())
        """
    ).strip()

    with (
        contextlib.closing(interpreters.create()) as interp1,
        contextlib.closing(interpreters.create()) as interp2,
    ):
        with pytest.raises(
            interpreters.ExecutionFailed,
            match="does not support loading in subinterpreters",
        ):
            interp1.exec("import mod_shared_interpreter_gil")

        values = interpreters.create_queue()
        interp1.prepare_main(values=values)
        interp1.exec(code)
        res1 = values.get_nowait()

        interp2.prepare_main(values=values)
        interp2.exec(code)
        res2 = values.get_nowait()

    assert res1 != m.internals_at(), "internals should differ from main interpreter"
    assert res2 != m.internals_at(), "internals should differ from main interpreter"
    assert res1 != res2, "internals should differ between interpreters"


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
def test_dependent_subinterpreters():
    """Makes sure the internals object differs across subinterpreters"""

    sys.path.insert(0, os.path.dirname(pybind11_tests.__file__))

    run_string, create = get_interpreters(modern=False)

    import mod_shared_interpreter_gil as m

    if not m.defined_PYBIND11_HAS_SUBINTERPRETER_SUPPORT:
        pytest.skip("Does not have subinterpreter support compiled in")

    code = textwrap.dedent(
        """
        import mod_shared_interpreter_gil as m
        import pickle
        with open(pipeo, 'wb') as f:
            pickle.dump(m.internals_at(), f)
        """
    ).strip()

    with create("legacy") as interp1:
        pipei, pipeo = os.pipe()
        run_string(interp1, code, shared={"pipeo": pipeo})
        with open(pipei, "rb") as f:
            res1 = pickle.load(f)

    assert res1 != m.internals_at(), "internals should differ from main interpreter"


PREAMBLE_CODE = textwrap.dedent(
    f"""
    def test():
        import sys

        sys.path.insert(0, {os.path.dirname(env.__file__)!r})
        sys.path.insert(0, {os.path.dirname(pybind11_tests.__file__)!r})

        import collections
        import mod_per_interpreter_gil_with_singleton as m

        objects = m.get_objects_in_singleton()
        expected = [
            type(None),               # static type: shared between interpreters
            tuple,                    # static type: shared between interpreters
            list,                     # static type: shared between interpreters
            dict,                     # static type: shared between interpreters
            collections.OrderedDict,  # static type: shared between interpreters
            collections.defaultdict,  # heap type: dynamically created per interpreter
            collections.deque,        # heap type: dynamically created per interpreter
        ]
        # Check that we have the expected objects. Avoid IndexError by checking lengths first.
        assert len(objects) == len(expected), (
            f"Expected {{expected!r}} ({{len(expected)}}), got {{objects!r}} ({{len(objects)}})."
        )
        # The first ones are static types shared between interpreters.
        assert objects[:-2] == expected[:-2], (
            f"Expected static objects {{expected[:-2]!r}}, got {{objects[:-2]!r}}."
        )
        # The last two are heap types created per-interpreter.
        # The expected objects are dynamically imported from `collections`.
        assert objects[-2:] == expected[-2:], (
            f"Expected heap objects {{expected[-2:]!r}}, got {{objects[-2:]!r}}."
        )

        assert hasattr(m, 'MyClass'), "Module missing MyClass"
        assert hasattr(m, 'MyGlobalError'), "Module missing MyGlobalError"
        assert hasattr(m, 'MyLocalError'), "Module missing MyLocalError"
        assert hasattr(m, 'MyEnum'), "Module missing MyEnum"
    """
).lstrip()


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
@pytest.mark.skipif(not CONCURRENT_INTERPRETERS_SUPPORT, reason="Requires 3.14.0b3+")
def test_import_module_with_singleton_per_interpreter():
    """Tests that a singleton storing Python objects works correctly per-interpreter"""
    from concurrent import interpreters

    code = f"{PREAMBLE_CODE.strip()}\n\ntest()\n"
    with contextlib.closing(interpreters.create()) as interp:
        interp.exec(code)


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
@pytest.mark.skipif(not CONCURRENT_INTERPRETERS_SUPPORT, reason="Requires 3.14.0b3+")
def test_import_in_subinterpreter_after_main():
    """Tests that importing a module in a subinterpreter after the main interpreter works correctly"""
    env.check_script_success_in_subprocess(
        PREAMBLE_CODE
        + textwrap.dedent(
            """
            import contextlib
            import gc
            from concurrent import interpreters

            test()

            interp = None
            with contextlib.closing(interpreters.create()) as interp:
                interp.call(test)

            del interp
            for _ in range(5):
                gc.collect()
            """
        )
    )

    env.check_script_success_in_subprocess(
        PREAMBLE_CODE
        + textwrap.dedent(
            """
            import contextlib
            import gc
            import random
            from concurrent import interpreters

            test()

            interps = interp = None
            with contextlib.ExitStack() as stack:
                interps = [
                    stack.enter_context(contextlib.closing(interpreters.create()))
                    for _ in range(8)
                ]
                random.shuffle(interps)
                for interp in interps:
                    interp.call(test)

            del interps, interp, stack
            for _ in range(5):
                gc.collect()
            """
        )
    )


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
@pytest.mark.skipif(not CONCURRENT_INTERPRETERS_SUPPORT, reason="Requires 3.14.0b3+")
def test_import_in_subinterpreter_before_main():
    """Tests that importing a module in a subinterpreter before the main interpreter works correctly"""
    env.check_script_success_in_subprocess(
        PREAMBLE_CODE
        + textwrap.dedent(
            """
            import contextlib
            import gc
            from concurrent import interpreters

            interp = None
            with contextlib.closing(interpreters.create()) as interp:
                interp.call(test)

            test()

            del interp
            for _ in range(5):
                gc.collect()
            """
        )
    )

    env.check_script_success_in_subprocess(
        PREAMBLE_CODE
        + textwrap.dedent(
            """
            import contextlib
            import gc
            from concurrent import interpreters

            interps = interp = None
            with contextlib.ExitStack() as stack:
                interps = [
                    stack.enter_context(contextlib.closing(interpreters.create()))
                    for _ in range(8)
                ]
                for interp in interps:
                    interp.call(test)

            test()

            del interps, interp, stack
            for _ in range(5):
                gc.collect()
            """
        )
    )

    env.check_script_success_in_subprocess(
        PREAMBLE_CODE
        + textwrap.dedent(
            """
            import contextlib
            import gc
            from concurrent import interpreters

            interps = interp = None
            with contextlib.ExitStack() as stack:
                interps = [
                    stack.enter_context(contextlib.closing(interpreters.create()))
                    for _ in range(8)
                ]
                for interp in interps:
                    interp.call(test)

                test()

            del interps, interp, stack
            for _ in range(5):
                gc.collect()
            """
        )
    )


@pytest.mark.skipif(
    sys.platform.startswith("emscripten"), reason="Requires loadable modules"
)
@pytest.mark.xfail(
    env.MUSLLINUX,
    reason="Flaky on musllinux, see also: https://github.com/pybind/pybind11/pull/5972#discussion_r2755283335",
    strict=False,
)
@pytest.mark.skipif(not CONCURRENT_INTERPRETERS_SUPPORT, reason="Requires 3.14.0b3+")
def test_import_in_subinterpreter_concurrently():
    """Tests that importing a module in multiple subinterpreters concurrently works correctly"""
    env.check_script_success_in_subprocess(
        PREAMBLE_CODE
        + textwrap.dedent(
            """
            import gc
            from concurrent.futures import InterpreterPoolExecutor, as_completed

            futures = future = None
            with InterpreterPoolExecutor(max_workers=16) as executor:
                futures = [executor.submit(test) for _ in range(32)]
                for future in as_completed(futures):
                    future.result()
            del futures, future, executor

            for _ in range(5):
                gc.collect()
            """
        )
    )



from __future__ import annotations

import pytest

from pybind11_tests import class_sh_mi_thunks as m


def test_ptrdiff_drvd_base0():
    ptrdiff = m.ptrdiff_drvd_base0()
    # A failure here does not (necessarily) mean that there is a bug, but that
    # test_class_sh_mi_thunks is not exercising what it is supposed to.
    # If this ever fails on some platforms: use pytest.skip()
    # If this ever fails on all platforms: don't know, seems extremely unlikely.
    assert ptrdiff != 0


@pytest.mark.parametrize(
    "vec_size_fn",
    [
        m.vec_size_base0_raw_ptr,
        m.vec_size_base0_shared_ptr,
    ],
)
@pytest.mark.parametrize(
    "get_fn",
    [
        m.get_drvd_as_base0_raw_ptr,
        m.get_drvd_as_base0_shared_ptr,
        m.get_drvd_as_base0_unique_ptr,
    ],
)
def test_get_vec_size_raw_shared(get_fn, vec_size_fn):
    obj = get_fn()
    assert vec_size_fn(obj) == 5


@pytest.mark.parametrize(
    "get_fn", [m.get_drvd_as_base0_raw_ptr, m.get_drvd_as_base0_unique_ptr]
)
def test_get_vec_size_unique(get_fn):
    obj = get_fn()
    assert m.vec_size_base0_unique_ptr(obj) == 5
    with pytest.raises(ValueError, match="Python instance was disowned"):
        m.vec_size_base0_unique_ptr(obj)


def test_get_shared_vec_size_unique():
    obj = m.get_drvd_as_base0_shared_ptr()
    with pytest.raises(ValueError) as exc_info:
        m.vec_size_base0_unique_ptr(obj)
    assert (
        str(exc_info.value) == "Cannot disown external shared_ptr (load_as_unique_ptr)."
    )


def test_virtual_base_not_at_offset_0():
    # This test ensures that the Diamond fixture actually exercises a non-zero
    # virtual-base subobject offset on our supported platforms/ABIs.
    #
    # If this assert ever fails on some platform/toolchain, please adjust the
    # C++ fixture so the virtual base is *not* at offset 0:
    #   - Keep VBase non-empty.
    #   - Make Left and Right non-empty and asymmetrically sized and, if
    #     needed, nudge with a modest alignment.
    #   - The goal is to achieve a non-zero address delta between `Diamond*`
    #     and `static_cast<VBase*>(Diamond*)`.
    #
    # Rationale: certain smart_holder features are exercised only when the
    # registered subobject address differs from the most-derived object start,
    # so this check guards test efficacy across compilers.
    addrs = m.diamond_addrs()
    assert addrs.as_vbase - addrs.as_self != 0, (
        "Diamond VBase at offset 0 on this platform; to ensure test efficacy, "
        "tweak fixtures (VBase/Left/Right) to ensure non-zero subobject offset."
    )


@pytest.mark.parametrize(
    "make_fn",
    [
        m.make_diamond_as_vbase_raw_ptr,  # exercises smart_holder::from_raw_ptr_take_ownership
        m.make_diamond_as_vbase_shared_ptr,  # exercises smart_holder_from_shared_ptr
        m.make_diamond_as_vbase_unique_ptr,  # exercises smart_holder_from_unique_ptr
    ],
)
def test_make_diamond_as_vbase(make_fn):
    # Added under PR #5836
    vb = make_fn()
    assert vb.ping() == 7


@pytest.mark.parametrize(
    "clone_fn",
    [
        m.Tiger.clone_raw_ptr,
        m.Tiger.clone_shared_ptr,
        m.Tiger.clone_unique_ptr,
    ],
)
def test_animal_cat_tiger(clone_fn):
    # Based on Animal-Cat-Tiger reproducer under PR #5796
    tiger = m.Tiger()
    cloned = clone_fn(tiger)
    assert isinstance(cloned, m.Tiger)



from __future__ import annotations

import os
import re
import sys

DIR = os.path.abspath(os.path.dirname(__file__))

# shlex/sysconfig are imported lazily below to keep `import pybind11` light

# This is the conditional used for os.path being posixpath
if "posix" in sys.builtin_module_names:

    def _quote(s: str) -> str:
        import shlex

        return shlex.quote(s)
elif "nt" in sys.builtin_module_names:
    # See https://github.com/mesonbuild/meson/blob/db22551ed9d2dd7889abea01cc1c7bba02bf1c75/mesonbuild/utils/universal.py#L1092-L1121
    # and the original documents:
    # https://docs.microsoft.com/en-us/cpp/c-language/parsing-c-command-line-arguments and
    # https://blogs.msdn.microsoft.com/twistylittlepassagesallalike/2011/04/23/everyone-quotes-command-line-arguments-the-wrong-way/
    _UNSAFE = re.compile("[ \t\n\r]")

    def _quote(s: str) -> str:
        if s and not _UNSAFE.search(s):
            return s

        # Paths cannot contain a '"' on Windows, so we don't need to worry
        # about nuanced counting here.
        return f'"{s}\\"' if s.endswith("\\") else f'"{s}"'
else:

    def _quote(s: str) -> str:
        return s


def _config(name: str) -> str:
    """A stripped sysconfig variable, or "" if unset."""
    import sysconfig

    return str(sysconfig.get_config_var(name) or "").strip()


def get_include(user: bool = False) -> str:  # noqa: ARG001
    """
    Return the path to the pybind11 include directory. The historical "user"
    argument is unused, and may be removed.
    """
    installed_path = os.path.join(DIR, "include")
    source_path = os.path.join(os.path.dirname(DIR), "include")
    return installed_path if os.path.exists(installed_path) else source_path


def get_cmake_dir() -> str:
    """
    Return the path to the pybind11 CMake module directory.
    """
    cmake_installed_path = os.path.join(DIR, "share", "cmake", "pybind11")
    if os.path.exists(cmake_installed_path):
        return cmake_installed_path

    msg = "pybind11 not installed, installation required to access the CMake files"
    raise ImportError(msg)


def get_pkgconfig_dir() -> str:
    """
    Return the path to the pybind11 pkgconfig directory.
    """
    pkgconfig_installed_path = os.path.join(DIR, "share", "pkgconfig")
    if os.path.exists(pkgconfig_installed_path):
        return pkgconfig_installed_path

    msg = "pybind11 not installed, installation required to access the pkgconfig files"
    raise ImportError(msg)


def get_include_dirs() -> list[str]:
    """
    Return the unique include directories for Python and pybind11.
    """
    import sysconfig

    dirs = [
        sysconfig.get_path("include"),
        sysconfig.get_path("platinclude"),
        get_include(),
    ]

    # Make unique but preserve order
    unique_dirs = []
    for d in dirs:
        if d and d not in unique_dirs:
            unique_dirs.append(d)
    return unique_dirs


def get_cflags() -> str:
    """
    Return the compile flags for building a simple extension with a
    Unix-style command-line compiler (GCC/Clang). Based on python-config.
    """
    flags = [_quote(f"-I{d}") for d in get_include_dirs()]
    # CFLAGS is a pre-composed multi-flag string, passed through as-is (like
    # python-config does)
    if cflags := _config("CFLAGS"):
        flags.append(cflags)
    flags.append("-std=c++17")
    return " ".join(flags)


def get_ldflags(embed: bool = False) -> str:
    """
    Return the link flags for building a simple extension (or, with
    embed=True, an embedding program) with a Unix-style command-line
    compiler (GCC/Clang). Based on python-config.
    """
    # LDFLAGS/LIBS/SYSLIBS are pre-composed multi-flag strings, passed
    # through as-is (like python-config does)
    flags = [ldflags] if (ldflags := _config("LDFLAGS")) else []

    if embed:
        if libdir := _config("LIBDIR"):
            flags.append(_quote(f"-L{libdir}"))
        if not _config("Py_ENABLE_SHARED") and (libpl := _config("LIBPL")):
            flags.append(_quote(f"-L{libpl}"))
        abiflags = getattr(sys, "abiflags", "") or ""
        flags.append(f"-lpython{_config('VERSION')}{abiflags}")
        if libs := _config("LIBS"):
            flags.append(libs)
        if syslibs := _config("SYSLIBS"):
            flags.append(syslibs)
    elif sys.platform.startswith("darwin"):
        flags += ["-undefined", "dynamic_lookup", "-shared"]
    elif os.name == "posix":
        # Linux and other Unix platforms (FreeBSD, Solaris, AIX, ...)
        flags += ["-fPIC", "-shared"]

    return " ".join(flags)



"""pytest configuration

Extends output capture as needed by pybind11: ignore constructors, optional unordered lines.
Adds docstring and exceptions message sanitizers.
"""

from __future__ import annotations

import contextlib
import difflib
import gc
import importlib.metadata
import multiprocessing
import re
import sys
import sysconfig
import textwrap
import traceback
import weakref
from typing import Callable

import pytest

# Early diagnostic for failed imports
try:
    import pybind11_tests
except Exception:
    # pytest does not show the traceback without this.
    traceback.print_exc()
    raise


@pytest.fixture(scope="session", autouse=True)
def use_multiprocessing_forkserver_on_linux():
    if sys.platform != "linux" or sys.implementation.name == "graalpy":
        # The default on Windows, macOS and GraalPy is "spawn": If it's not broken, don't fix it.
        return

    # Full background: https://github.com/pybind/pybind11/issues/4105#issuecomment-1301004592
    # In a nutshell: fork() after starting threads == flakiness in the form of deadlocks.
    # It is actually a well-known pitfall, unfortunately without guard rails.
    # "forkserver" is more performant than "spawn" (~9s vs ~13s for tests/test_gil_scoped.py,
    # visit the issuecomment link above for details).
    multiprocessing.set_start_method("forkserver")


_long_marker = re.compile(r"([0-9])L")
_hexadecimal = re.compile(r"0x[0-9a-fA-F]+")

# Avoid collecting Python3 only files
collect_ignore = []


def _strip_and_dedent(s):
    """For triple-quote strings"""
    return textwrap.dedent(s.lstrip("\n").rstrip())


def _split_and_sort(s):
    """For output which does not require specific line order"""
    return sorted(_strip_and_dedent(s).splitlines())


def _make_explanation(a, b):
    """Explanation for a failed assert -- the a and b arguments are List[str]"""
    return ["--- actual / +++ expected"] + [
        line.strip("\n") for line in difflib.ndiff(a, b)
    ]


class Output:
    """Basic output post-processing and comparison"""

    def __init__(self, string):
        self.string = string
        self.explanation = []

    def __str__(self):
        return self.string

    __hash__ = None

    def __eq__(self, other):
        # Ignore constructor/destructor output which is prefixed with "###"
        a = [
            line
            for line in self.string.strip().splitlines()
            if not line.startswith("###")
        ]
        b = _strip_and_dedent(other).splitlines()
        if a == b:
            return True
        self.explanation = _make_explanation(a, b)
        return False


class Unordered(Output):
    """Custom comparison for output without strict line ordering"""

    __hash__ = None

    def __eq__(self, other):
        a = _split_and_sort(self.string)
        b = _split_and_sort(other)
        if a == b:
            return True
        self.explanation = _make_explanation(a, b)
        return False


class Capture:
    def __init__(self, capfd):
        self.capfd = capfd
        self.out = ""
        self.err = ""

    def __enter__(self):
        self.capfd.readouterr()
        return self

    def __exit__(self, *args):
        self.out, self.err = self.capfd.readouterr()

    __hash__ = None

    def __eq__(self, other):
        a = Output(self.out)
        b = other
        if a == b:
            return True
        self.explanation = a.explanation
        return False

    def __str__(self):
        return self.out

    def __contains__(self, item):
        return item in self.out

    @property
    def unordered(self):
        return Unordered(self.out)

    @property
    def stderr(self):
        return Output(self.err)


@pytest.fixture
def capture(capsys):
    """Extended `capsys` with context manager and custom equality operators"""
    return Capture(capsys)


class SanitizedString:
    def __init__(self, sanitizer):
        self.sanitizer = sanitizer
        self.string = ""
        self.explanation = []

    def __call__(self, thing):
        self.string = self.sanitizer(thing)
        return self

    __hash__ = None

    def __eq__(self, other):
        a = self.string
        b = _strip_and_dedent(other)
        if a == b:
            return True
        self.explanation = _make_explanation(a.splitlines(), b.splitlines())
        return False


def _sanitize_general(s):
    s = s.strip()
    s = s.replace("pybind11_tests.", "m.")
    return _long_marker.sub(r"\1", s)


def _sanitize_docstring(thing):
    s = thing.__doc__
    return _sanitize_general(s)


@pytest.fixture
def doc():
    """Sanitize docstrings and add custom failure explanation"""
    return SanitizedString(_sanitize_docstring)


def _sanitize_message(thing):
    s = str(thing)
    s = _sanitize_general(s)
    return _hexadecimal.sub("0", s)


@pytest.fixture
def msg():
    """Sanitize messages and add custom failure explanation"""
    return SanitizedString(_sanitize_message)


def pytest_assertrepr_compare(op, left, right):  # noqa: ARG001
    """Hook to insert custom failure explanation"""
    if hasattr(left, "explanation"):
        return left.explanation
    return None


# Number of times we think repeatedly collecting garbage might do anything.
# The only reason to do more than once is because finalizers executed during
# one GC pass could create garbage that can't be collected until a future one.
# This quickly produces diminishing returns, and GC passes can be slow, so this
# value is a tradeoff between non-flakiness and fast tests. (It errs on the
# side of non-flakiness; many uses of this idiom only do 3 passes.)
num_gc_collect = 5


def gc_collect():
    """Run the garbage collector several times (needed when running
    reference counting tests with PyPy)"""
    for _ in range(num_gc_collect):
        gc.collect()


def delattr_and_ensure_destroyed(*specs):
    """For each of the given *specs* (a tuple of the form ``(scope, name)``),
    perform ``delattr(scope, name)``, then do enough GC collections that the
    deleted reference has actually caused the target to be destroyed. This is
    typically used to test what happens when a type object is destroyed; if you
    use it for that, you should be aware that extension types, or all types,
    are immortal on some Python versions. See ``env.TYPES_ARE_IMMORTAL``.
    """
    wrs = []
    for mod, name in specs:
        wrs.append(weakref.ref(getattr(mod, name)))
        delattr(mod, name)

    for _ in range(num_gc_collect):
        gc.collect()
        if all(wr() is None for wr in wrs):
            break
    else:
        # If this fires, most likely something is still holding a reference
        # to the object you tried to destroy - for example, it's a type that
        # still has some instances alive. Try setting a breakpoint here and
        # examining `gc.get_referrers(wrs[0]())`. It's vaguely possible that
        # num_gc_collect needs to be increased also.
        pytest.fail(
            f"Could not delete bindings such as {next(wr for wr in wrs if wr() is not None)!r}"
        )


def pytest_configure():
    pytest.suppress = contextlib.suppress
    pytest.gc_collect = gc_collect
    pytest.delattr_and_ensure_destroyed = delattr_and_ensure_destroyed


def pytest_report_header():
    assert pybind11_tests.compiler_info is not None, (
        "Please update pybind11_tests.cpp if this assert fails."
    )
    interesting_packages = ("pybind11", "numpy", "scipy", "build")
    valid = []
    for package in sorted(interesting_packages):
        with contextlib.suppress(ModuleNotFoundError):
            valid.append(f"{package}=={importlib.metadata.version(package)}")
    reqs = " ".join(valid)

    cpp_info = [
        "C++ Info:",
        f"{pybind11_tests.compiler_info}",
        f"{pybind11_tests.cpp_std}",
        f"{pybind11_tests.PYBIND11_INTERNALS_ID}",
        f"PYBIND11_SIMPLE_GIL_MANAGEMENT={pybind11_tests.PYBIND11_SIMPLE_GIL_MANAGEMENT}",
    ]
    if "__graalpython__" in sys.modules:
        cpp_info.append(
            f"GraalPy version: {sys.modules['__graalpython__'].get_graalvm_version()}"
        )
    lines = [
        f"installed packages of interest: {reqs}",
        " ".join(cpp_info),
    ]
    if sysconfig.get_config_var("Py_GIL_DISABLED"):
        lines.append("free-threaded Python build")

    return lines


@pytest.fixture
def backport_typehints() -> Callable[[SanitizedString], SanitizedString]:
    d = {}
    if sys.version_info < (3, 13):
        d["typing_extensions.TypeIs"] = "typing.TypeIs"
        d["typing_extensions.CapsuleType"] = "types.CapsuleType"
    if sys.version_info < (3, 12):
        d["typing_extensions.Buffer"] = "collections.abc.Buffer"
    if sys.version_info < (3, 11):
        d["typing_extensions.Never"] = "typing.Never"
    if sys.version_info < (3, 10):
        d["typing_extensions.TypeGuard"] = "typing.TypeGuard"

    def backport(sanitized_string: SanitizedString) -> SanitizedString:
        for old, new in d.items():
            sanitized_string.string = sanitized_string.string.replace(old, new)

        return sanitized_string

    return backport



from __future__ import annotations

import os

import env


def test_enum_import_exit_no_crash():
    # Added in PR #6015. Modeled after reproducer under issue #5976
    env.check_script_success_in_subprocess(
        f"""
        import sys
        sys.path.insert(0, {os.path.dirname(env.__file__)!r})
        import standalone_enum_module as m
        assert m.SomeEnum.__class__.__name__ == "pybind11_type"
        """,
        rerun=1,
    )



from __future__ import annotations

import contextlib
import importlib
import os
import re
import shutil
import subprocess
import sys
import sysconfig
import tarfile
import zipfile
from collections.abc import Generator
from pathlib import Path

import pytest

# These tests must be run explicitly

DIR = Path(__file__).parent.resolve()
MAIN_DIR = DIR.parent.parent

FILENAME_VERSION = re.compile(r"[-_]((\d+\.\d+\.\d+)(?:[a-z]+\d*)?)(?:-|\.tar\.gz$)")

# Newer pytest has global path setting, but keeping old pytest for now
sys.path.append(str(MAIN_DIR / "tools"))

from make_global import get_global

HAS_UV = shutil.which("uv") is not None
UV_ARGS = ["--installer=uv"] if HAS_UV else []

PKGCONFIG = """\
prefix=${{pcfiledir}}/../../
includedir=${{prefix}}/include

Name: pybind11
Description: Seamless operability between C++11 and Python
Version: {VERSION}
Cflags: -I${{includedir}}
"""


main_headers = {
    "include/pybind11/attr.h",
    "include/pybind11/buffer_info.h",
    "include/pybind11/cast.h",
    "include/pybind11/chrono.h",
    "include/pybind11/common.h",
    "include/pybind11/complex.h",
    "include/pybind11/critical_section.h",
    "include/pybind11/eigen.h",
    "include/pybind11/embed.h",
    "include/pybind11/eval.h",
    "include/pybind11/functional.h",
    "include/pybind11/gil.h",
    "include/pybind11/gil_safe_call_once.h",
    "include/pybind11/gil_simple.h",
    "include/pybind11/iostream.h",
    "include/pybind11/native_enum.h",
    "include/pybind11/numpy.h",
    "include/pybind11/operators.h",
    "include/pybind11/options.h",
    "include/pybind11/pybind11.h",
    "include/pybind11/pytypes.h",
    "include/pybind11/subinterpreter.h",
    "include/pybind11/stl.h",
    "include/pybind11/stl_bind.h",
    "include/pybind11/trampoline_self_life_support.h",
    "include/pybind11/type_caster_pyobject_ptr.h",
    "include/pybind11/typing.h",
    "include/pybind11/warnings.h",
}

conduit_headers = {
    "include/pybind11/conduit/README.txt",
    "include/pybind11/conduit/pybind11_conduit_v1.h",
    "include/pybind11/conduit/pybind11_platform_abi_id.h",
    "include/pybind11/conduit/wrap_include_python_h.h",
}

detail_headers = {
    "include/pybind11/detail/argument_vector.h",
    "include/pybind11/detail/class.h",
    "include/pybind11/detail/common.h",
    "include/pybind11/detail/cpp_conduit.h",
    "include/pybind11/detail/descr.h",
    "include/pybind11/detail/dynamic_raw_ptr_cast_if_possible.h",
    "include/pybind11/detail/function_record_pyobject.h",
    "include/pybind11/detail/function_ref.h",
    "include/pybind11/detail/holder_caster_foreign_helpers.h",
    "include/pybind11/detail/init.h",
    "include/pybind11/detail/internals.h",
    "include/pybind11/detail/native_enum_data.h",
    "include/pybind11/detail/pybind11_namespace_macros.h",
    "include/pybind11/detail/struct_smart_holder.h",
    "include/pybind11/detail/type_caster_base.h",
    "include/pybind11/detail/typeid.h",
    "include/pybind11/detail/using_smart_holder.h",
    "include/pybind11/detail/value_and_holder.h",
    "include/pybind11/detail/exception_translation.h",
}

eigen_headers = {
    "include/pybind11/eigen/common.h",
    "include/pybind11/eigen/matrix.h",
    "include/pybind11/eigen/tensor.h",
}

stl_headers = {
    "include/pybind11/stl/filesystem.h",
}

cmake_files = {
    "share/cmake/pybind11/FindPythonLibsNew.cmake",
    "share/cmake/pybind11/pybind11Common.cmake",
    "share/cmake/pybind11/pybind11Config.cmake",
    "share/cmake/pybind11/pybind11ConfigVersion.cmake",
    "share/cmake/pybind11/pybind11GuessPythonExtSuffix.cmake",
    "share/cmake/pybind11/pybind11NewTools.cmake",
    "share/cmake/pybind11/pybind11Targets.cmake",
    "share/cmake/pybind11/pybind11Tools.cmake",
}

pkgconfig_files = {
    "share/pkgconfig/pybind11.pc",
}

py_files = {
    "__init__.py",
    "__main__.py",
    "_version.py",
    "commands.py",
    "py.typed",
    "setup_helpers.py",
    "share/__init__.py",
    "share/pkgconfig/__init__.py",
}

headers = main_headers | conduit_headers | detail_headers | eigen_headers | stl_headers
generated_files = cmake_files | pkgconfig_files
all_files = headers | generated_files | py_files

sdist_files = {
    "pyproject.toml",
    "LICENSE",
    "README.rst",
    "PKG-INFO",
    "SECURITY.md",
}


@contextlib.contextmanager
def preserve_file(filename: Path) -> Generator[str, None, None]:
    old_stat = filename.stat()
    old_file = filename.read_text(encoding="utf-8")
    try:
        yield old_file
    finally:
        filename.write_text(old_file, encoding="utf-8")
        os.utime(filename, (old_stat.st_atime, old_stat.st_mtime))


@contextlib.contextmanager
def build_global() -> Generator[None, None, None]:
    """
    Build global SDist and wheel.
    """

    pyproject = MAIN_DIR / "pyproject.toml"
    with preserve_file(pyproject):
        newer_txt = get_global()
        pyproject.write_text(newer_txt, encoding="utf-8")
        yield


def read_tz_file(tar: tarfile.TarFile, name: str) -> bytes:
    start = tar.getnames()[0].split("/")[0] + "/"
    inner_file = tar.extractfile(tar.getmember(f"{start}{name}"))
    assert inner_file
    with contextlib.closing(inner_file) as f:
        return f.read()


def normalize_line_endings(value: bytes) -> bytes:
    return value.replace(os.linesep.encode("utf-8"), b"\n")


def test_build_sdist(monkeypatch, tmpdir):
    monkeypatch.chdir(MAIN_DIR)

    subprocess.run(
        [sys.executable, "-m", "build", "--sdist", f"--outdir={tmpdir}", *UV_ARGS],
        check=True,
    )

    (sdist,) = tmpdir.visit("*.tar.gz")
    version = FILENAME_VERSION.search(sdist.basename).group(1)

    with tarfile.open(str(sdist), "r:gz") as tar:
        simpler = {n.split("/", 1)[-1] for n in tar.getnames()[1:]}
        (pkg_info_path,) = (n for n in simpler if n.endswith("PKG-INFO"))

        pyproject_toml = read_tz_file(tar, "pyproject.toml")
        pkg_info = read_tz_file(tar, pkg_info_path).decode("utf-8")

    files = headers | sdist_files
    assert files <= simpler

    assert b'name = "pybind11"' in pyproject_toml
    assert f"Version: {version}" in pkg_info
    assert "License-Expression: BSD-3-Clause" in pkg_info
    assert "License-File: LICENSE" in pkg_info
    assert "Provides-Extra: global" in pkg_info
    assert f'Requires-Dist: pybind11-global=={version}; extra == "global"' in pkg_info


def test_build_global_dist(monkeypatch, tmpdir):
    monkeypatch.chdir(MAIN_DIR)
    with build_global():
        subprocess.run(
            [
                sys.executable,
                "-m",
                "build",
                "--sdist",
                "--outdir",
                str(tmpdir),
                *UV_ARGS,
            ],
            check=True,
        )

    (sdist,) = tmpdir.visit("*.tar.gz")
    version = FILENAME_VERSION.search(sdist.basename).group(2)

    with tarfile.open(str(sdist), "r:gz") as tar:
        simpler = {n.split("/", 1)[-1] for n in tar.getnames()[1:]}
        (pkg_info_path,) = (n for n in simpler if n.endswith("PKG-INFO"))

        pyproject_toml = read_tz_file(tar, "pyproject.toml")
        pkg_info = read_tz_file(tar, pkg_info_path).decode("utf-8")

    files = headers | sdist_files
    assert files <= simpler

    assert b'name = "pybind11-global"' in pyproject_toml
    assert f"Version: {version}" in pkg_info
    assert "License-Expression: BSD-3-Clause" in pkg_info
    assert "License-File: LICENSE" in pkg_info
    assert "Provides-Extra: global" not in pkg_info
    assert 'Requires-Dist: pybind11-global; extra == "global"' not in pkg_info


def tests_build_wheel(monkeypatch, tmpdir):
    monkeypatch.chdir(MAIN_DIR)

    subprocess.run(
        [sys.executable, "-m", "build", "--wheel", "--outdir", str(tmpdir), *UV_ARGS],
        check=True,
    )

    (wheel,) = tmpdir.visit("*.whl")
    version, simple_version = FILENAME_VERSION.search(wheel.basename).groups()

    files = {f"pybind11/{n}" for n in all_files}
    files |= {
        "dist-info/licenses/LICENSE",
        "dist-info/METADATA",
        "dist-info/RECORD",
        "dist-info/WHEEL",
        "dist-info/entry_points.txt",
    }

    with zipfile.ZipFile(str(wheel)) as z:
        names = z.namelist()
        share = zipfile.Path(z, "pybind11/share")
        pkgconfig = (share / "pkgconfig/pybind11.pc").read_text(encoding="utf-8")
        cmakeconfig = (share / "cmake/pybind11/pybind11Config.cmake").read_text(
            encoding="utf-8"
        )
        (pkg_info_path,) = (n for n in names if n.endswith("METADATA"))
        pkg_info = zipfile.Path(z, pkg_info_path).read_text(encoding="utf-8")

    trimmed = {n for n in names if "dist-info" not in n}
    trimmed |= {f"dist-info/{n.split('/', 1)[-1]}" for n in names if "dist-info" in n}

    assert files == trimmed

    assert 'set(pybind11_INCLUDE_DIR "${PACKAGE_PREFIX_DIR}/include")' in cmakeconfig

    pkgconfig_expected = PKGCONFIG.format(VERSION=simple_version)
    assert pkgconfig_expected == pkgconfig

    assert f"Version: {version}" in pkg_info
    assert "License-Expression: BSD-3-Clause" in pkg_info
    assert "License-File: LICENSE" in pkg_info
    assert "Provides-Extra: global" in pkg_info
    assert f'Requires-Dist: pybind11-global=={version}; extra == "global"' in pkg_info


def tests_build_global_wheel(monkeypatch, tmpdir):
    monkeypatch.chdir(MAIN_DIR)
    with build_global():
        subprocess.run(
            [
                sys.executable,
                "-m",
                "build",
                "--wheel",
                "--outdir",
                str(tmpdir),
                *UV_ARGS,
            ],
            check=True,
        )

    (wheel,) = tmpdir.visit("*.whl")
    version, simple_version = FILENAME_VERSION.search(wheel.basename).groups()

    files = {f"data/data/{n}" for n in headers}
    files |= {f"data/headers/{n[8:]}" for n in headers}
    files |= {f"data/data/{n}" for n in generated_files}
    files |= {
        "dist-info/licenses/LICENSE",
        "dist-info/METADATA",
        "dist-info/WHEEL",
        "dist-info/RECORD",
    }

    with zipfile.ZipFile(str(wheel)) as z:
        names = z.namelist()
        beginning = names[0].split("/", 1)[0].rsplit(".", 1)[0]

        share = zipfile.Path(z, f"{beginning}.data/data/share")
        pkgconfig = (share / "pkgconfig/pybind11.pc").read_text(encoding="utf-8")
        cmakeconfig = (share / "cmake/pybind11/pybind11Config.cmake").read_text(
            encoding="utf-8"
        )

        (pkg_info_path,) = (n for n in names if n.endswith("METADATA"))
        pkg_info = zipfile.Path(z, pkg_info_path).read_text(encoding="utf-8")

    assert f"Version: {version}" in pkg_info
    assert "License-Expression: BSD-3-Clause" in pkg_info
    assert "License-File: LICENSE" in pkg_info
    assert "Provides-Extra: global" not in pkg_info
    assert 'Requires-Dist: pybind11-global; extra == "global"' not in pkg_info

    trimmed = {n[len(beginning) + 1 :] for n in names}

    assert files == trimmed

    assert 'set(pybind11_INCLUDE_DIR "${PACKAGE_PREFIX_DIR}/include")' in cmakeconfig

    pkgconfig_expected = PKGCONFIG.format(VERSION=simple_version)
    assert pkgconfig_expected == pkgconfig


def test_version_matches():
    header = MAIN_DIR / "include/pybind11/detail/common.h"
    text = header.read_text()

    # Extract the relevant macro values
    regex_prefix = r"#\s*define\s+PYBIND11_VERSION_"
    micro = re.search(rf"{regex_prefix}MICRO\s+(\d+)\b", text).group(1)
    release_level = re.search(rf"{regex_prefix}RELEASE_LEVEL\s+(\w+)\b", text).group(1)
    release_serial = re.search(
        rf"{regex_prefix}RELEASE_SERIAL\s+(\d+)\b",
        text,
    ).group(1)
    patch = re.search(rf"{regex_prefix}PATCH\s+([\w.-]+)\b", text).group(1)

    # Map release level macro to string
    level_map = {
        "PY_RELEASE_LEVEL_ALPHA": "a",
        "PY_RELEASE_LEVEL_BETA": "b",
        "PY_RELEASE_LEVEL_GAMMA": "rc",
        "PY_RELEASE_LEVEL_FINAL": "",
    }
    level_str = level_map[release_level]

    if release_level == "PY_RELEASE_LEVEL_FINAL":
        assert level_str == ""
        assert release_serial == "0"
        expected_patch = micro
    else:
        expected_patch = f"{micro}{level_str}{release_serial}"

    assert patch == expected_patch


def run_command_line(*args: str) -> str:
    env = os.environ.copy()
    env["PYTHONPATH"] = str(MAIN_DIR)
    result = subprocess.run(
        [sys.executable, "-m", "pybind11", *args],
        capture_output=True,
        text=True,
        check=True,
        env=env,
    )
    return result.stdout


def test_cli_cflags():
    out = run_command_line("--cflags")
    assert "-std=c++17" in out
    assert f"-I{sysconfig.get_path('include')}" in out


def test_cli_ldflags_embed():
    out = run_command_line("--ldflags", "--embed")
    assert "-lpython" in out
    if sysconfig.get_config_var("LIBDIR"):
        assert "-L" in out


@pytest.mark.skipif(os.name != "posix", reason="quote style is platform-specific")
def test_cflags_quotes_paths_with_spaces(monkeypatch):
    monkeypatch.syspath_prepend(str(MAIN_DIR))
    commands = importlib.import_module("pybind11.commands")
    monkeypatch.setattr(sysconfig, "get_path", lambda name: f"/spa ced/{name}")
    assert "'-I/spa ced/include'" in commands.get_cflags()


@pytest.mark.skipif(os.name != "posix", reason="Unix link flags only")
def test_ldflags_other_unix(monkeypatch):
    monkeypatch.syspath_prepend(str(MAIN_DIR))
    commands = importlib.import_module("pybind11.commands")
    monkeypatch.setattr(sys, "platform", "freebsd14")
    monkeypatch.setattr(commands, "_config", lambda name: "")  # noqa: ARG005
    out = commands.get_ldflags()
    assert "-shared" in out
    assert "-fPIC" in out


def test_cli_file():
    out = run_command_line("--file", "example.cpp").rstrip()
    ext_suffix = sysconfig.get_config_var("EXT_SUFFIX")
    assert "-std=c++17" in out
    assert out.index("-std=c++17") < out.index("example.cpp")
    assert out.endswith(f"-o example{ext_suffix}")
    if sys.platform.startswith(("linux", "darwin")):
        assert out.index("example.cpp") < out.index("-shared")


def test_cli_file_embed():
    out = run_command_line("--file", "example.cpp", "--embed").rstrip()
    assert out.endswith("-o example")



from __future__ import annotations

import types

import pytest

import env
from pybind11_tests import class_cross_module_use_after_one_module_dealloc as m


@pytest.mark.skipif(
    env.TYPES_ARE_IMMORTAL, reason="can't GC type objects on this platform"
)
def test_cross_module_use_after_one_module_dealloc():
    # This is a regression test for a bug that occurred during development of
    # internals::registered_types_cpp_fast (see #5842). registered_types_cpp_fast maps
    # &typeid(T) to a raw non-owning pointer to a Python type object. If two DSOs both
    # look up the same global type, they will create two separate entries in
    # registered_types_cpp_fast, which will look like:
    # +=========================================+
    # |&typeid(T) from DSO 1|type object pointer|
    # |&typeid(T) from DSO 2|type object pointer|
    # +=========================================+
    #
    # Then, if the type object is destroyed and we don't take extra steps to clean up
    # the table thoroughly, the first row of the table will be cleaned up but the second
    # one will contain a dangling pointer to the old type object. Further lookups from
    # DSO 2 will then return that dangling pointer, which will cause use-after-frees.

    import pybind11_cross_module_tests as cm

    module_scope = types.ModuleType("module_scope")
    instance = m.register_and_instantiate_cross_dso_class(module_scope)
    cm.consume_cross_dso_class(instance)

    del instance
    pytest.delattr_and_ensure_destroyed((module_scope, "CrossDSOClass"))

    # Make sure that CrossDSOClass gets allocated at a different address.
    m.register_unrelated_class(module_scope)

    instance = m.register_and_instantiate_cross_dso_class(module_scope)
    cm.consume_cross_dso_class(instance)



# pylint: disable=missing-function-docstring
from __future__ import annotations

import argparse
import functools
import sys
import sysconfig
from pathlib import Path

from ._version import __version__
from .commands import (
    _quote as quote,
)
from .commands import (
    get_cflags,
    get_cmake_dir,
    get_include_dirs,
    get_ldflags,
    get_pkgconfig_dir,
)


def print_includes() -> None:
    print(" ".join(quote(f"-I{d}") for d in get_include_dirs()))


def main() -> None:
    make_parser = functools.partial(argparse.ArgumentParser, allow_abbrev=False)
    if sys.version_info >= (3, 14):
        make_parser = functools.partial(make_parser, color=True, suggest_on_error=True)
    parser = make_parser()
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print the version and exit.",
    )
    parser.add_argument(
        "--includes",
        action="store_true",
        help="Include flags for both pybind11 and Python headers.",
    )
    parser.add_argument(
        "--cmakedir",
        action="store_true",
        help="Print the CMake module directory, ideal for setting -Dpybind11_ROOT in CMake.",
    )
    parser.add_argument(
        "--pkgconfigdir",
        action="store_true",
        help="Print the pkgconfig directory, ideal for setting $PKG_CONFIG_PATH.",
    )
    parser.add_argument(
        "--extension-suffix",
        action="store_true",
        help="Print the extension for a Python module",
    )
    parser.add_argument(
        "--cflags",
        action="store_true",
        help="Print the compile flags for a simple extension (Unix-style compilers).",
    )
    parser.add_argument(
        "--ldflags",
        action="store_true",
        help="Print the link flags for a simple extension (Unix-style compilers).",
    )
    parser.add_argument(
        "--embed",
        action="store_true",
        help="Build for embedding instead of an extension; affects --ldflags and --file.",
    )
    parser.add_argument(
        "--file",
        type=Path,
        help="Print a full command-line suffix for compiling the given file;"
        " the output goes next to the source file (Unix-style compilers).",
    )
    args = parser.parse_args()
    if not sys.argv[1:]:
        parser.print_help()
    ext_suffix = sysconfig.get_config_var("EXT_SUFFIX") or ""
    if args.file:
        suffix = "" if args.embed else ext_suffix
        print(
            get_cflags(),
            quote(str(args.file)),
            get_ldflags(embed=args.embed),
            "-o",
            quote(str(args.file.with_suffix(suffix))),
        )
    else:
        if args.cflags:
            print(get_cflags())
        if args.ldflags:
            print(get_ldflags(embed=args.embed))
    # --cflags and --file already contain the include flags
    if args.includes and not (args.cflags or args.file):
        print_includes()
    if args.cmakedir:
        print(quote(get_cmake_dir()))
    if args.pkgconfigdir:
        print(quote(get_pkgconfig_dir()))
    if args.extension_suffix:
        print(ext_suffix)


if __name__ == "__main__":
    main()



from __future__ import annotations

import gc
import os
import weakref

import pytest

import env
import pybind11_tests
from pybind11_tests import custom_type_setup as m


@pytest.fixture
def gc_tester():
    """Tests that an object is garbage collected.

    Assumes that any unreferenced objects are fully collected after calling
    `gc.collect()`.  That is true on CPython, but does not appear to reliably
    hold on PyPy.
    """

    weak_refs = []

    def add_ref(obj):
        # PyPy does not support `gc.is_tracked`.
        if hasattr(gc, "is_tracked"):
            assert gc.is_tracked(obj)
        weak_refs.append(weakref.ref(obj))

    yield add_ref

    gc.collect()
    for ref in weak_refs:
        assert ref() is None


# PyPy does not seem to reliably garbage collect.
@pytest.mark.skipif("env.PYPY or env.GRAALPY")
def test_self_cycle(gc_tester):
    obj = m.ContainerOwnsPythonObjects()
    obj.append(obj)
    gc_tester(obj)


# PyPy does not seem to reliably garbage collect.
@pytest.mark.skipif("env.PYPY or env.GRAALPY")
def test_indirect_cycle(gc_tester):
    obj = m.ContainerOwnsPythonObjects()
    obj.append([obj])
    gc_tester(obj)


@pytest.mark.skipif("env.PYPY or env.GRAALPY")
def test_py_cast_useable_on_shutdown():
    """Test that py::cast works during interpreter shutdown.

    See PR #5972 and https://github.com/pybind/pybind11/pull/5958#discussion_r2717645230.
    """
    env.check_script_success_in_subprocess(
        f"""
        import sys

        sys.path.insert(0, {os.path.dirname(env.__file__)!r})
        sys.path.insert(0, {os.path.dirname(pybind11_tests.__file__)!r})

        from pybind11_tests import custom_type_setup as m

        # Create a self-referential cycle that will be collected during shutdown.
        # The tp_traverse and tp_clear callbacks call py::cast, which requires
        # internals to still be valid.
        obj = m.ContainerOwnsPythonObjects()
        obj.append(obj)

        # Add weakref callbacks that verify the capsule is still alive when the
        # pybind11 object is garbage collected during shutdown.
        m.add_gc_checkers_with_weakrefs(obj)
        """
    )



#!/usr/bin/env -S uv run

# /// script
# dependencies = ["ghapi>=2", "rich"]
# ///

from __future__ import annotations

import os
import re
import subprocess

import ghapi.all
from rich import print
from rich.syntax import Syntax

MD_ENTRY = re.compile(
    r"""
    ^\#+\ Suggested\ changelog\ entry:?\s*$ # Match the heading, colon optional
    (?:\s*<!--.*?-->)?                     # Optionally match one comment
    (?P<content>.*?)                       # Lazily capture content until...
    (?=                                    # Lookahead for one of the following:
        ^-{3,}\s*$                         #   A line with 3 or more dashes
      | ^<!--\s*readthedocs                #   A comment starting with 'readthedocs'
      | ^\#\#                              #   A new heading
      | \Z                                 #   End of string
    )
    """,
    re.DOTALL | re.VERBOSE | re.MULTILINE,
)
# Conventional commit prefix, such as "fix(cmake)!:"
TITLE_CAT = re.compile(r"(?P<cat>\w+)(?:\((?P<sub>[^)]+)\))?!?:")

print()


def get_token() -> str | None:
    """
    Unauthenticated requests get a shared CDN cache ("Cache-Control: public"),
    which returns stale bodies and labels. A token makes the cache private.
    """
    token = os.environ.get("GITHUB_TOKEN") or os.environ.get("GH_TOKEN")
    if token:
        return token
    try:
        result = subprocess.run(
            ["gh", "auth", "token"], capture_output=True, text=True, check=True
        )
    except (OSError, subprocess.CalledProcessError):
        print("[yellow]No GitHub token found; results may be stale (cached).")
        return None
    return result.stdout.strip()


LABEL = "needs changelog"

api = ghapi.all.GhApi(owner="pybind", repo="pybind11", token=get_token(), sync=True)

issues_pages = ghapi.page.sync_paged(
    api.issues.list_for_repo, labels=LABEL, state="closed", per_page=100
)
# The server-side label filter lags behind label removals, so re-check each issue
issues = (
    issue
    for page in issues_pages
    for issue in page
    if any(label.name == LABEL for label in issue.labels)
)
missing = []
old = []
cats_descr = {
    "feat": "New Features",
    "fix": "Bug fixes",
    "docs": "Documentation",
    "tests": "Tests",
    "ci": "CI",
    "chore": "Other",
    "unknown": "Uncategorised",
}
# Each main category maps its subcategories ("" if there is none) to entries
cats: dict[str, dict[str, list[str]]] = {c: {} for c in cats_descr}

for issue in issues:
    if "```rst" in issue.body:
        old.append(issue)
        continue

    changelog = MD_ENTRY.search(issue.body or "")
    if not changelog:
        missing.append(issue)
        continue

    msg = changelog.group("content").strip()
    if not msg:
        missing.append(issue)
        continue
    msg = msg.removeprefix("* ")
    if not msg.startswith("- "):
        msg = "- " + msg
    if not msg.endswith("."):
        msg += "."
    if msg == "- Placeholder.":
        missing.append(issue)
        continue

    msg += f"\n  [#{issue.number}]({issue.html_url})"
    title = TITLE_CAT.match(issue.title)
    cat = title.group("cat").lower() if title else "unknown"
    sub = (title.group("sub") or "").lower() if title else ""
    if cat not in cats:
        cat, sub = "unknown", ""
    cats[cat].setdefault(sub, []).append(msg)

for cat, subs in cats.items():
    if subs:
        print(f"[bold]{cats_descr[cat]}:")
        print()
        # An empty subcategory sorts first, so plain entries lead the section
        for sub in sorted(subs):
            if sub:
                print(f"<!-- {cat}({sub}) -->")
                print()
            for msg in subs[sub]:
                print(Syntax(msg, "md", theme="ansi_light", word_wrap=True))
                print()
            print()

if missing:
    print()
    print("[blue]" + "-" * 30)
    print()

    for issue in missing:
        print(f"[red bold]Missing:[/red bold][red] {issue.title}")
        print(f"[red]  {issue.html_url}\n")

    print("[bold]Template:\n")
    msg = "## Suggested changelog entry:"
    print(Syntax(msg, "md", theme="ansi_light"))

if old:
    print()
    print("[red]" + "-" * 30)
    print()

    for issue in old:
        print(f"[red bold]Old:[/red bold][red] {issue.title}")
        print(f"[red]  {issue.html_url}\n")

print()



from __future__ import annotations

from pybind11_tests import pybind11_pytorch_regressions as m


def test_pytorch_like_get_tracing_state_aliases_singleton_shared_ptr():
    a = m._get_tracing_state()
    b = m._get_tracing_state()

    a.value = 17

    assert b.value == 17
    assert m._get_tracing_state().value == 17


def test_pytorch_like_compilation_unit_get_interface_aliases_member_shared_ptr():
    cu = m.CompilationUnit()

    a = cu.get_interface("iface")
    b = cu.get_interface("iface")

    a.value = 23

    assert b.value == 23
    assert cu.get_interface("iface").value == 23



from __future__ import annotations

import platform
import sys
import sysconfig

import pytest

ANDROID = sys.platform.startswith("android")
IOS = sys.platform.startswith("ios")
LINUX = sys.platform.startswith("linux")
MACOS = sys.platform.startswith("darwin")
WIN = sys.platform.startswith("win32") or sys.platform.startswith("cygwin")
FREEBSD = sys.platform.startswith("freebsd")

MUSLLINUX = False
MANYLINUX = False
if LINUX:

    def _is_musl() -> bool:
        libc, _ = platform.libc_ver()
        return libc == "musl" or (libc != "glibc" and libc != "")

    MUSLLINUX = _is_musl()
    MANYLINUX = not MUSLLINUX
    del _is_musl

CPYTHON = platform.python_implementation() == "CPython"
PYPY = platform.python_implementation() == "PyPy"
PYPY_PRE_7_3_23 = PYPY and sys.pypy_version_info < (7, 3, 23)
GRAALPY = sys.implementation.name == "graalpy"

# Compile-time config (what the binary was built for)
PY_GIL_DISABLED = bool(sysconfig.get_config_var("Py_GIL_DISABLED"))
# Runtime state (what's actually happening now)
sys_is_gil_enabled = getattr(sys, "_is_gil_enabled", lambda: True)

TYPES_ARE_IMMORTAL = (
    PYPY
    or GRAALPY
    or (CPYTHON and PY_GIL_DISABLED and (3, 13) <= sys.version_info < (3, 14))
)


def check_script_success_in_subprocess(code: str, *, rerun: int = 8) -> None:
    """Runs the given code in a subprocess."""
    import os
    import subprocess
    import sys
    import textwrap

    if ANDROID or IOS or sys.platform.startswith("emscripten"):
        pytest.skip("Requires subprocess support")

    code = textwrap.dedent(code).strip()
    try:
        for _ in range(rerun):  # run flakily failing test multiple times
            subprocess.check_output(
                [sys.executable, "-c", code],
                cwd=os.getcwd(),
                stderr=subprocess.STDOUT,
                text=True,
            )
    except subprocess.CalledProcessError as ex:
        raise RuntimeError(
            f"Subprocess failed with exit code {ex.returncode}.\n\n"
            f"Code:\n"
            f"```python\n"
            f"{code}\n"
            f"```\n\n"
            f"Output:\n"
            f"{ex.output}"
        ) from None
